class EventHook(object):
    def __init__(self):
        self.event_handlers = []

    def __iadd__(self, handler):
        self.event_handlers.append(handler)
        return self

    def __isub__(self, handler):
        self.event_handlers.remove(handler)
        return self

    def fire(self, *args, **keywargs):
        for handler in self.event_handlers:
            handler(*args, **keywargs)



import logging
# import coloredlogs

# LEVEL = logging.DEBUG
LEVEL = logging.INFO

FORMAT = '%(asctime)-15s : %(levelname)-8s : %(message)s'
logging.basicConfig(format=FORMAT)
# logging.addLevelName(logging.DEBUG, 'DBG')
# logging.addLevelName(logging.DEBUG, 'ERR')
logger = logging.getLogger('ipc')
logger.level = LEVEL
# handler = logging.StreamHandler()
# logger.addHandler(handler)
# handler.setFormatter(ColoredFormatter())

# coloredlogs.install(level=LEVEL, logger=logger)

debug = logger.debug
info = logger.info
warning = logger.warning
warn = logger.warn
error = logger.error

def set_level(level):
    logger.level = level



import pika
import traceback
import sys
from threading import Thread, Event

from isc import codecs, log

from time import sleep


class Node(object):
    """
    Registers services & listens to AMQP for RPC calls & notifications.
    """

    def __init__(self, hostname='127.0.0.1', url=None, exchange='isc'):
        self.exchange = exchange
        self._infix = '_service_'
        self.queue_name_offset = len(exchange) + len(self._infix)
        self.services = {}
        self.listeners = {}
        self._is_ready = Event()
        if url is None:
            self.params = pika.ConnectionParameters(hostname)
        else:
            self.params = pika.URLParameters(url)
        self._is_running = False
        self.codecs = {}
        self.register_codec(codecs.PickleCodec())
        self.register_codec(codecs.JSONCodec())
        self.hooks = {
            'pre_call': set(),
            'post_success': set(),
            'post_error': set()
        }

    def run(self):
        """
        Starts provider node.
        Blocks until `stop` is called.
        """
        self._is_running = True
        while self._is_running:
            try:
                self.conn = pika.BlockingConnection(self.params)
                self.conn.process_data_events = self._fix_pika_timeout(self.conn.process_data_events)
                self.channel = self.conn.channel()

                self._create_service_queues(self.channel, self.services)
                self._register_listeners(self.services)
                self._create_fanout_exchange(self.channel)
                self._schedule_methods(self.services)

                log.info('Ready')
                self._is_ready.set()
                self.channel.start_consuming()
                self._is_running = False
                self._is_ready.clear()
            except pika.exceptions.ConnectionClosed:  # pragma: no cover
                self._is_ready.clear()
                log.error('Connection closed, retrying in 3 seconds')
                sleep(3)
                continue
        log.info('Node terminated')

    def register_service(self, service):
        """
        Registers a service to make it callable & notifiable via RPC.
        """
        if service.name in self.services:
            raise Exception('Service {} is already registered.'.format(service.name))
        self.services[service.name] = service

    def register_codec(self, codec):
        """
        Registers a codec for this server.
        Codec must be an instance of :class:`isc.codecs.AbstractCodec` subclass.
        Doesn't remove previously registered codecs.
        """
        self.codecs[codec.content_type] = codec

    def add_hook(self, name):
        """
        Registers a hook to be executed when requested event occurs.
        Possible names are `pre_call`, `post_success` or `post_error`.
        """
        def decorator(fn):
            self.hooks[name] |= set([fn])
            return fn
        return decorator

    def wait_for_ready(self):
        """
        Blocks until the connection to AMQP is established.
        """
        self._is_ready.wait()

    def stop(self):
        """
        Stops provider node.
        """
        self.conn.add_timeout(0, lambda: self.channel.stop_consuming())

    def _fix_pika_timeout(self, process_data_events):
        """
        Fixes pika timeout (default is 5, we want it to be 1 for faster shutdown)
        """
        def process_data_events_new(time_limit=0):
            return process_data_events(time_limit=1)
        return process_data_events_new

    def _run_scheduled_with_local_timer(self, fn, timeout):
        """
        Runs the method and reschedules it to be executed again.
        """
        fn_name = fn.__name__
        exception, _ = self._call_service_method(fn, (), {})
        if not exception:
            log.debug('Scheduled function %s completed successfully.', fn_name)
        self._schedule_with_local_timer(fn, timeout)

    def _schedule_with_local_timer(self, fn, timeout):
        """
        Schedules a job to be executed after requested timeout.
        """
        Thread(
            target=self._wait_and_run,
            args=(
                timeout,
                self._run_scheduled_with_local_timer,
                (fn, timeout),
                {}
            )
        ).start()

    def _wait_and_run(self, interval, fn, args, kwargs):
        sleep(interval)
        fn(*args, **kwargs)

    def _create_service_queues(self, channel, services):
        """
        Creates necessary AMQP queues, one per service.
        """
        channel.exchange_declare(exchange=self.exchange)
        for service in services.values():
            queue = '{}_service_{}'.format(self.exchange, service.name)
            channel.queue_declare(queue=queue)
            channel.queue_bind(queue, self.exchange)
            channel.basic_consume(self._on_message, queue=queue, no_ack=False)

    def _register_listeners(self, services):
        """
        Populates listeners dictionary to speed up notification handling.
        """
        for service in services.values():
            for attr in [getattr(service, attr, None) for attr in dir(service)]:
                for event in getattr(attr, '__on__', []):
                    if event not in self.listeners:
                        self.listeners[event] = []
                    self.listeners[event].append(attr)

    def _create_fanout_exchange(self, channel):
        """
        Creates a fanout queue to accept notifications.
        """
        channel.exchange_declare(exchange='{}_fanout'.format(self.exchange), type='fanout')
        fanout_queue = channel.queue_declare(exclusive=True)
        channel.queue_bind(exchange='{}_fanout'.format(self.exchange), queue=fanout_queue.method.queue)

        channel.basic_consume(self._on_broadcast, queue=fanout_queue.method.queue, no_ack=True)

    def _schedule_methods(self, services):
        """
        Spawns timers for periodic tasks.
        """
        for service in services.values():
            for attr in [getattr(service, attr, None) for attr in dir(service)]:
                for timeout in getattr(attr, '__timeouts__', []):
                    self._schedule_with_local_timer(attr, timeout)

    def _on_message(self, channel, method, properties, body):
        """
        Called when a message is received on one of the queues.
        """
        Thread(target=self._validate_message, args=(channel, method, properties, body)).start()

    def _validate_message(self, channel, method, properties, body):
        """
        Checks and acknowledges a received message if it can be handled
        by any registered service.
        """

        service_name = method.routing_key[self.queue_name_offset:]

        if service_name not in self.services:  # pragma: no cover
            return

        channel.basic_ack(delivery_tag=method.delivery_tag)

        # service = self.services[service_name]
        try:
            log.debug('Got invocation ..{}'.format(
                str(properties.correlation_id)[-4:]
            ))
            requested_codec, (fn_name, args, kwargs) = self._decode_message(properties, body)
        except Exception as e:
            log.error(str(e))
        else:
            result = self._call_service_method((service_name, fn_name), args, kwargs)

            log.debug('Publishing response for invocation ..{}'.format(
                str(properties.correlation_id)[-4:]
            ))

            channel.basic_publish(exchange=self.exchange, routing_key=properties.reply_to, properties=pika.BasicProperties(
                correlation_id=properties.correlation_id
            ), body=requested_codec.encode(result))

    def _decode_message(self, properties, body):
        """
        Decodes message body.
        Raises `Exception` on error.
        """
        content_type = properties.content_type

        try:
            codec = self.codecs[content_type]
        except Exception as e:
            raise Exception('Unknown codec: {}'.format(content_type))

        try:
            return codec, codec.decode(body)
        except Exception as e:
            raise Exception('Failed to decode message: {}'.format(str(e)))

    def _call_service_method(self, info, args, kwargs):
        """
        Calls method and returns the tuple with `Exception` instance
        or `None` and result or `None`.
        `info` can be tuple containing service name & method name
        or the actual callable method.
        """
        try:
            if isinstance(info, tuple):
                service_name, fn_name = info
                fn = self._get_method(service_name, fn_name)
            else:
                fn_name = info.__name__
                fn = info
            self._fire_hook('pre_call', fn_name, args, kwargs)
            result = (None, fn(*args, **kwargs))

            args_str = ', '.join(map(str, args))
            kwargs_str = ', '.join(['{}={}'.format(k, v) for k, v in kwargs.items()])

            args_kwargs_str = ', '.join((args_str, kwargs_str))

            log.debug('{}({})'.format(fn_name, args_kwargs_str))
            self._fire_hook('post_success', fn_name, args, kwargs, result)
            return result
        except Exception as e:
            self._log_method_error(fn_name, e)
            self._fire_hook('post_error', fn_name, args, kwargs, e)
            return (str(e), None)

    def _log_method_error(self, fn_name, e):
        """
        Logs error with appropriate stack frame extracted.
        """
        tb = sys.exc_info()[2]
        frame = traceback.extract_tb(tb)[-1]
        if isinstance(frame, tuple):  # pragma: no cover
            filename, lineno, _, line = frame
        else:  # pragma: no cover
            filename, lineno, line = frame.filename, frame.lineno, frame.line
        log.error('Error in RPC method "{}", file {}:{}:\n    {}\n{}: {}'.format(fn_name, filename, lineno, line, e.__class__.__name__, str(e)))

    def _get_method(self, service_name, fn_name):
        """
        Checks if requested method exists and can be called.
        Raises `Exception` on error.
        """
        service = self.services[service_name]
        fn = getattr(service, fn_name, None)

        if fn is None:
            log.warning('No method %s', fn_name)
            raise Exception('Could not find method {}.'.format(fn_name))

        if getattr(fn, '__exposed__', None) is None:
            log.warning('Method %s is not exposed', fn_name)
            raise Exception('You are not allowed to call unexposed method {}.'.format(fn_name))

        return fn

    def _on_broadcast(self, channel, method, properties, body):
        """
        Called when a notification is received.
        Does not acknowledge notifications because they're
        delivered via `fanout` exchange.
        """
        try:
            requested_codec, (event, data) = self._decode_message(properties, body)
        except Exception as e:
            log.error(str(e))
        else:
            log.debug('Got notification ..{}'.format(
                str(properties.correlation_id)[-4:]
            ))

            listeners = self.listeners.get(event, [])
            for fn in listeners:
                Thread(target=fn, args=(data,)).start()

    def _fire_hook(self, name, *args, **kwargs):
        """
        Fires methods registered for a requested hook.
        """
        for hook in self.hooks[name]:
            hook(*args, **kwargs)

    def set_logging_level(self, level):
        """
        Set logging level.
        """
        log.set_level(level)


def expose(fn):
    """
    Marks a method as "exposed", i. e. available to be called via RPC.
    """
    fn.__exposed__ = True
    return fn


def on(event):
    """
    Marks a method as "event handler", i. e. reacting to notifications.
    """
    def wrapper(fn):
        if getattr(fn, '__on__', None) is None:
            fn.__on__ = set()
        fn.__on__ |= set([event])
        return fn
    return wrapper


def local_timer(timeout):
    """
    Marks a method as "periodic job" to execute it every `timeout` seconds.
    """
    def wrapper(fn):
        if getattr(fn, '__timeouts__', None) is None:
            fn.__timeouts__ = set()
        fn.__timeouts__ |= set([timeout])
        return fn
    return wrapper



import pickle
import json


class AbstractCodec(object):
    """
    Abstract base class for implementing codecs.

    "Codec" is a class that tells ISC how to encode & decode message payloads.

    Server can support multiple codecs while client can use only one at a time.

    Default codec is :class:`.PickleCodec`. You can implement your own codec
    by extending :class:`.AbstractCodec` and overriding its methods.

    Important: Don't forget that both client and server should have the codec installed!
    """
    def encode(self, message):  # pragma: no cover
        """
        Called when a message needs to be serialized
        for sending over AMQP channel.
        """
        raise NotImplementedError()

    def decode(self, payload):  # pragma: no cover
        """
        Called when a message needs to be serialized
        for sending over AMQP channel.
        """
        raise NotImplementedError()


class PickleCodec(AbstractCodec):
    """
    Pickle codec implementation.
    """
    content_type = 'pickle'

    def encode(self, message):
        return pickle.dumps(message, protocol=2)

    def decode(self, payload):
        return pickle.loads(payload)


class JSONCodec(AbstractCodec):
    """
    JSON codec implementation.
    """
    content_type = 'json'

    def encode(self, message):
        return json.dumps(message)

    def decode(self, payload):
        return json.loads(payload)



from django.core.management import BaseCommand
from django.conf import settings
from isc.server import Node
from importlib import import_module


class Command(BaseCommand):
    def _import_object(self, import_string):
        module_name, _, object_name = import_string.rpartition('.')
        module = import_module(module_name)
        return getattr(module, object_name)

    def handle(self, *args, **kwargs):
        assert getattr(settings, 'ISC', None), 'ISC config not present in settings'
        assert 'hostname' in settings.ISC, 'hostname not provided in ISC config'
        assert 'services' in settings.ISC, 'services not provided in ISC config'

        node = Node(hostname=settings.ISC['hostname'])

        for service_import_string in settings.ISC['services']:
            service_class = self._import_object(service_import_string)
            node.register_service(service_class())

        for hook_name, hook_import_string in settings.ISC.get('hooks', {}).items():
            hook_function = self._import_object(hook_import_string)
            node.add_hook(hook_name)(hook_function)

        node.run()



import kombu
import uuid
import socket
from threading import Thread, Event
from six.moves.queue import Queue, Empty
from time import sleep
from weakref import proxy

from . import codecs, log
from .eventhook import EventHook


class IPCException(Exception):
    pass


class RemoteException(IPCException):
    pass


class LocalException(IPCException):
    pass


class TimeoutException(LocalException):
    pass


class FutureResult(object):
    """
    Encapsulates future result.
    Provides interface to block until future data is ready.
    Thread-safe.
    """
    def __init__(self, cannonical_name, **extra):
        self.event = Event()
        self.exception = None
        self.value = None
        self.cannonical_name = cannonical_name
        self.extra = extra

    def wait(self, timeout=5):
        """
        Blocks until data is ready.
        """
        if not self.event.wait(timeout=timeout):
            self.exception = TimeoutException()
            return False
        return True

    def resolve(self, value):
        """
        Resolves this promise with result and sets "ready" event.
        """
        if not self.is_ready():
            self.exception = None
            self.value = value
            self.event.set()

    def reject(self, exception):
        """
        Rejects this promise with exception and sets "ready" event.
        """
        if not self.is_ready():
            self.exception = exception
            self.value = None
            self.event.set()

    def is_ready(self):
        """
        Checks if this result has been resolved or rejected.
        """
        return self.event.is_set()


class QueuedRequest(object):
    """
    Internal class. Represents pending outgoing message.
    """
    def __init__(self, codec, **kwargs):
        self.codec = codec
        self.correlation_id = str(uuid.uuid4())
        self.__dict__.update(**kwargs)


class QueuedInvocation(QueuedRequest):
    """
    Internal class. Represents pending outgoing method call.
    """
    def __init__(self, codec, service, method, args, kwargs):
        super(QueuedInvocation, self).__init__(
            codec,
            service=service,
            method=method,
            args=args,
            kwargs=kwargs
        )


class QueuedNotification(QueuedRequest):
    """
    Internal class. Represents pending outgoing notification.
    """
    def __init__(self, codec, event, data):
        super(QueuedNotification, self).__init__(
            codec,
            event=event,
            data=data
        )


class ConsumerThread(Thread):
    """
    Internal class. Represents connection & message consuming thread.
    """

    def __init__(self, hostname, exchange_name, connect_timeout, reconnect_timeout, codec):
        super(ConsumerThread, self).__init__()

        self.daemon = True
        self._is_running = False
        self._is_connected = False

        self._hostname = hostname
        self._exchange_name = exchange_name

        self.on_connect = EventHook()
        self.on_error = EventHook()
        self.on_disconnect = EventHook()
        self.on_message = EventHook()

        self._connect_timeout = connect_timeout
        self._reconnect_timeout = reconnect_timeout

        if not isinstance(codec, codecs.AbstractCodec):
            codec = codecs.PickleCodec()
        self._codec = codec

    def run(self):
        self._is_running = True
        while self._is_running:
            try:
                log.info('Connecting to AMQP...')
                self._connect()
                self._is_connected = True
                self.on_connect.fire()
                log.info('Connected to AMQP')
                self._start_consuming()
                self._is_connected = False
                self.on_disconnect.fire()
                return
            except Exception as e:
                self._is_connected = False
                if self._reconnect_timeout:
                    log.error(
                        'Disconnected from AMQP.\n'
                        'Retrying in {} seconds.\n'
                        'Error was: {}'.format(
                            self._reconnect_timeout,
                            # format_exc()
                            str(e)
                        )
                    )
                    self.on_error.fire()
                    sleep(self._reconnect_timeout)
                    continue
                else:
                    log.error(
                        'Disconnected from AMQP.\n'
                        'Not reconnecting because reconnect_timeout = 0.\n'
                        'Error was: {}'.format(
                            # format_exc()
                            str(e)
                        )
                    )
                    self.on_error.fire()
                    return

    def _connect(self):
        self._conn = kombu.Connection(
            self._hostname,
            connect_timeout=self._connect_timeout
        )

        self._channel = self._conn.channel()

        self._exchange = kombu.Exchange(
            name=self._exchange_name,
            channel=self._channel,
            durable=False
        )

        self._callback_queue = self._create_callback_queue(
            self._channel,
            self._exchange
        )

    def _create_callback_queue(self, channel, exchange):
        name = 'response-{}'.format(uuid.uuid4())
        callback_queue = kombu.Queue(
            name=name,
            exchange=exchange,
            routing_key=name,
            exclusive=True,
            channel=self._channel
        )
        callback_queue.declare()
        return callback_queue

    def _start_consuming(self):
        """
        Start consuming messages.
        This function is blocking.
        """
        consumer = kombu.Consumer(
            self._conn,
            queues=[self._callback_queue],
            on_message=self._on_message,
            accept=[self._codec.content_type],
            no_ack=True
        )

        consumer.consume()

        while self._is_running:
            try:
                self._conn.drain_events(timeout=0.5)
            except socket.timeout:
                continue

    def _on_message(self, message):
        log.debug('Got response for invocation ..{}'.format(
            str(message.properties['correlation_id'])[-4:]
        ))
        self.on_message.fire(message)

    def is_connected(self):
        return self._is_connected

    def shutdown_worker(self):
        self._is_running = False

    def get_connection(self):
        return self._conn

    def get_codec(self):
        return self._codec

    def get_exchange(self):
        return self._exchange

    def get_callback_queue(self):
        return self._callback_queue


class PublisherThread(Thread):
    """
    Internal class. Represents message publishing thread.
    """

    def __init__(self, consumer):
        self.consumer = proxy(consumer)
        super(PublisherThread, self).__init__()
        self.daemon = True
        self._out_queue = Queue()
        self._is_running = False

    def run(self):
        self._is_running = True
        while self._is_running:
            if self.consumer.is_connected():
                try:
                    queued_request = self._out_queue.get(timeout=0.5)
                    with kombu.producers[self.consumer.get_connection()].acquire(block=True) as producer:
                        try:
                            self._dispatch_request(queued_request, producer)
                        except Exception as e:
                            # except ConnectionResetError:
                            log.debug('Failed to dispatch request, re-enqueueing again, error was: {}'.format(
                                str(e)
                            ))
                            self.enqueue(queued_request)
                except Empty:
                    continue
            else:
                sleep(0.5)
                log.debug('Waiting for consumer to be ready...')

    def enqueue(self, queued_request):
        self._out_queue.put(queued_request)

    def _dispatch_request(self, queued_request, producer):
        if isinstance(queued_request, QueuedInvocation):
            log.debug('Publishing invocation ..{}'.format(
                queued_request.correlation_id[-4:]
            ))
            producer.publish(
                exchange=self.consumer.get_exchange(),
                routing_key='{}_service_{}'.format(
                    self.consumer.get_exchange().name,
                    queued_request.service
                ),
                body=queued_request.codec.encode((
                    queued_request.method,
                    queued_request.args,
                    queued_request.kwargs
                )),
                reply_to=self.consumer.get_callback_queue().name,
                correlation_id=queued_request.correlation_id,
                content_type=self.consumer.get_codec().content_type,
            )
        else:
            log.debug('Publishing notification ..{}'.format(
                queued_request.correlation_id[-4:]
            ))
            producer.publish(
                exchange='{}_fanout'.format(
                    self.consumer.get_exchange().name
                ),
                routing_key='',
                body=queued_request.codec.encode((
                    queued_request.event,
                    queued_request.data
                )),
                correlation_id=queued_request.correlation_id,
                content_type=self.consumer.get_codec().content_type
            )

    def shutdown_worker(self):
        self._is_running = False


class Client(object):
    """
    Represents a single low-level connection to the ISC messaging broker.
    Thread-safe.
    """
    def __init__(self, host='amqp://guest:guest@127.0.0.1:5672/', exchange='isc', codec=None, connect_timeout=2, reconnect_timeout=3, invoke_timeout=20):
        self.future_results = {}

        self._invoke_timeout = invoke_timeout

        self._consumer = ConsumerThread(host, exchange, connect_timeout, reconnect_timeout, codec)
        self._publisher = PublisherThread(self._consumer)

        self._consumer.on_message += self._on_response

        self.on_connect = self._consumer.on_connect
        self.on_error = self._consumer.on_error
        self.on_disconnect = self._consumer.on_disconnect

    def start(self):
        """
        Start connection & publisher threads.
        """
        self._consumer.start()
        self._publisher.start()

    def stop(self):
        """
        Stops the client and waits for its termination.
        """
        self._publisher.shutdown_worker()
        self._consumer.shutdown_worker()
        self._publisher.join()
        self._consumer.join()

    def _on_response(self, message):
        """
        Called when a message is consumed.
        """
        future_result = self.future_results.get(message.properties['correlation_id'], None)
        if not future_result:  # pragma: no cover
            # TODO: Should not happen!
            log.error('FIXME: This should not happen.')
            return
        try:
            exception, result = self._consumer.get_codec().decode(message.body)
            if exception:
                exception = RemoteException(exception)
        except Exception as e:  # pragma: no cover
            exception, result = LocalException(str(e)), None

        if exception:
            future_result.reject(exception)
        else:
            future_result.resolve(result)

    def invoke_async(self, service, method, *args, **kwargs):
        """
        Serialize & publish method call request.
        """
        queued_request = QueuedInvocation(self._consumer.get_codec(), service, method, args, kwargs)

        future_result = FutureResult('{}.{}'.format(service, method))
        self.future_results[queued_request.correlation_id] = future_result

        self._publisher.enqueue(queued_request)

        return future_result

    def notify(self, event, data):
        """
        Serialize & publish notification.
        """
        queued_request = QueuedNotification(self._consumer.get_codec(), event, data)

        self._publisher.enqueue(queued_request)

    def invoke(self, service, method, *args, **kwargs):
        """
        Call a remote method and wait for a result.
        Blocks until a result is ready.
        """
        future_result = self.invoke_async(service, method, *args, **kwargs)

        future_result.wait(self._invoke_timeout)

        if future_result.exception:
            raise future_result.exception
        else:
            return future_result.value

    def set_invoke_timeout(self, timeout):
        """
        Sets timeout for waiting for results on this client.
        """
        self._invoke_timeout = timeout

    def set_logging_level(self, level):
        """
        Set logging level.
        """
        log.set_level(level)

    def __getattr__(self, attr):
        """
        Convenience method.
        Returns :class:`.ServiceProxy` to make it look like we're actually calling
        local methods from local objects.
        """
        return ServiceProxy(self, attr)


class ServiceProxy(object):
    """
    Convenience wrapper for service.

    It allows you to perform attribute chaining (e. g. :code:`client.example.add(2, 3)`)
    """
    def __init__(self, client, service_name):
        self.client = client
        self.service_name = service_name

    def __getattr__(self, attr):
        """
        Returns :class:`.MethodProxy`
        """
        return MethodProxy(self.client, self.service_name, attr)


class MethodProxy(object):
    """
    Convenience wrapper for method.

    It allows you to perform attribute chaining (e. g. :code:`client.example.add(2, 3)`)
    """
    def __init__(self, client, service_name, method_name):
        self.client = client
        self.service_name = service_name
        self.method_name = method_name

    def __call__(self, *args, **kwargs):
        """
        Finalizes the chain & performs actual RPC invocation.
        Blocks while waiting for result.

        Returns the result.

        This is same as calling :func:`~isc.client.Client.invoke`
        """
        return self.client.invoke(self.service_name, self.method_name, *args, **kwargs)

    def call_async(self, *args, **kwargs):
        """
        Finalizes the chain & performs actual RPC invocation.
        Does not block.

        Returns :class:`.FutureResult`.

        This is same as calling :func:`~isc.client.Client.invoke_async`
        """
        return self.client.invoke_async(self.service_name, self.method_name, *args, **kwargs)



from unittest import TestCase
try:  # pragma: no cover
    from unittest import mock
except ImportError:  # pragma: no cover
    import mock
from time import sleep
# from gevent import spawn
# from gevent.event import Event
from time import time
from .server import Node, expose, on, local_timer, log
from .client import Client, RemoteException, TimeoutException, FutureResult
from .codecs import JSONCodec
from threading import Thread, Event


class ExampleService(object):
    name = 'example'

    def __init__(self):
        # Just for the tests, don't do this - services *must* be stateless!
        self.stuff_done_event = Event()
        self.collect_stats_done = Event()

    @expose
    def add(self, a, b):
        return a + b

    @expose
    def raise_error(self):
        raise Exception('testing')

    def private_method(self):  # pragma: no cover
        return 'Cannot call me!'

    @on('boom')
    def do_stuff(self, arg):
        print('Got stuff:', arg)
        self.stuff_done_event.set()

    @expose
    def slow_method(self):  # pragma: no cover
        sleep(2)
        return 42

    @local_timer(timeout=1)
    def collect_stats(self):
        self.collect_stats_done.set()


class GenericTest(TestCase):
    def setUp(self):
        self.node = Node(exchange='isc-unittest')
        self.service = ExampleService()
        self.node.register_service(self.service)
        self.node_thread = Thread(target=self.node.run)
        self.node_thread.daemon = True
        self.node_thread.start()
        self.node.wait_for_ready()
        self.client = Client(exchange='isc-unittest')
        self.client.start()
        self.clients = []

    def tearDown(self):
        self.client.stop()
        self.node.stop()
        self.node_thread.join()
        for client in self.clients:
            client.stop()

    def test_method_success(self):
        self.assertEqual(self.client.example.add(2, 3), 5)
        self.assertEqual(self.client.invoke('example', 'add', 2, 3), 5)

    def test_method_async(self):
        future_result = self.client.example.add.call_async(2, 3)
        self.assertIsInstance(future_result, FutureResult)
        future_result.wait()
        self.assertEqual(future_result.exception, None)
        self.assertEqual(future_result.value, 5)

    def test_method_exceptions(self):
        self.assertRaises(RemoteException, self.client.example.add)
        self.assertRaises(RemoteException, self.client.example.add, (2, '3'))

    def test_unexisting_method(self):
        self.assertRaises(RemoteException, self.client.example.unexisting_method)
        # self.assertRaises(RemoteException, self.client.unexisting_service.add, (2, 3))

    def test_unexisting_service(self):
        self.client.set_invoke_timeout(1)
        self.assertRaises(TimeoutException, self.client.unexisting_service.some_method)

    def test_raises_exception(self):
        self.assertRaises(RemoteException, self.client.example.raise_error)
        try:
            self.client.example.raise_error()
        except RemoteException as e:
            self.assertEqual(str(e), 'testing')

    def test_private_method(self):
        self.assertRaises(RemoteException, self.client.example.private_method)

    def test_notify(self):
        self.client.notify('boom', dict(place='some_place'))
        self.assertEqual(self.service.stuff_done_event.wait(3), True)

    def test_register_again(self):
        self.assertRaises(Exception, self.node.register_service, self.service)

    def test_reconnect_and_redeliver(self):
        client = Client(host='amqp://127.0.0.1:55553', exchange='isc-unittest')
        client.start()
        self.clients.append(client)

        future_result = client.example.add.call_async(2, 3)

        error_event = Event()
        client.on_error += error_event.set
        self.assertTrue(error_event.wait(5))

        client._consumer._hostname = 'amqp://127.0.0.1:5672'

        connect_event = Event()
        client.on_connect += connect_event.set
        self.assertTrue(connect_event.wait(5))

        self.assertTrue(future_result.wait(5))

    def test_local_timer_timings(self):
        # Measure time taken by timer to execute.

        self.service.collect_stats_done.wait(timeout=5)

        timediffs = []
        samples = 5

        for i in range(samples):
            self.service.collect_stats_done.clear()
            start = time()
            self.service.collect_stats_done.wait()
            timediffs.append(time() - start)

        timediff_avg = sum(timediffs) / samples

        self.assertAlmostEqual(timediff_avg, 1, 1)

    # def test_bad_message_payload(self):
    #     e = Event()
    #     with mock.patch.object(self.client.codec, 'encode', return_value='crap'):
    #         with mock.patch.object(log, 'error', side_effect=lambda *args: e.set()) as error:
    #             self.client.example.add.call_async(2, 3)
    #             self.assertTrue(e.wait(1), 'log.error was not called')
    #             self.assertTrue(error.call_args[0][0].startswith('Failed to decode message'))

    #     self.assertEquals(self.client.example.add(2, 3), 5, 'Should operate normally after error')

    # def test_bad_notify_payload(self):
    #     e = Event()
    #     with mock.patch.object(self.client.codec, 'encode', return_value='crap'):
    #         with mock.patch.object(log, 'error', side_effect=lambda *args: e.set()) as error:
    #             self.client.notify('boom', dict(place='some_place'))
    #             self.assertTrue(e.wait(1), 'log.error was not called')
    #             self.assertTrue(error.call_args[0][0].startswith('Failed to decode message'))

    #     self.client.notify('boom', dict(place='some_place'))
    #     self.assertEqual(self.service.stuff_done_event.wait(3), True, 'Should operate normally after error')

    def test_slow_method(self):
        self.client.set_invoke_timeout(1)
        self.assertRaises(TimeoutException, self.client.example.slow_method)
        self.client.set_invoke_timeout(3)
        self.assertEqual(self.client.example.slow_method(), 42)

    def test_pickle_codec(self):
        self.assertEqual(self.client.example.add((2,), (3,)), (2, 3), 'When using pickle codec, tuple should not be downgraded to list.')

    def test_json_codec(self):
        client = Client(codec=JSONCodec(), exchange='isc-unittest')
        client.start()
        self.clients.append(client)
        self.assertEqual(client.example.add((2,), (3,)), [2, 3], 'When using JSON codec, tuple should be downgraded to list.')

    def test_no_reconnect(self):
        client = Client('amqp://127.0.0.1:55553', exchange='unexisting-exchange-this-will-fail', reconnect_timeout=0)
        self.clients.append(client)

        connect_event = Event()
        client.on_connect += connect_event.set
        error_event = Event()
        client.on_error += error_event.set

        client.start()

        sleep(3)

        self.assertFalse(connect_event.is_set())
        self.assertTrue(error_event.wait(3))

    def test_eventhook(self):
        client = Client(exchange='isc-unittest')
        self.clients.append(client)

        event = Event()

        client.on_connect += event.set

        client.start()
        self.assertTrue(event.wait(5))

        client.on_connect -= event.set



#!/usr/bin/env python3.6

# coding=utf-8

from setuptools import setup

setup(
    name='isclib',
    version='0.24',
    description='ISC: Inter-service communication layer for Python. Compatible with gevent.',
    author="Andrew Dunai",
    author_email='andrew@dun.ai',
    url='https://github.com/and3rson/isc',
    license='GPLv3',
    packages=[
        'isc',
        'isc.management',
        'isc.management.commands',
    ],
    include_package_data=True,
    install_requires=['setuptools', 'pika', 'kombu'],
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    keywords='rpc,python2,python3,python,gevent,amqp,pika,kombu',
)
